import xml.etree.ElementTree as ET
import argparse
import os

##parses the directory path to allow the file to be renamed
def file_prefixer(name,output,prefix):
    (head,tail)=os.path.split(name)
    tail=prefix+tail
    return os.path.join(output,tail)

##Finds the index of an XML element in a list of elements
##Finds first appearence on default, but can look for later ones
def indexGet(elements, tag, instance=0):
    elem=elements.findall(tag)
    index=elements.getchildren().index(elem[instance])
    return index

class CommentedTreeBuilder(ET.XMLTreeBuilder ):
    def __init__ ( self, html = 0, target = None ):
        ET.XMLTreeBuilder.__init__( self, html, target )
        self._parser.CommentHandler = self.handle_comment
    
    def handle_comment(self, data):
        self._target.start(ET.Comment, {})
        self._target.data(data)
        self._target.end(ET.Comment)

class recipeMaker():
    x=[str(y) for y in range (1,8)]
    w=[str(z) for z in range (9,28)]
    deadVolume=dict.fromkeys(x,15)
    chemNames=["FirstBase", "CompleteCycle", "EndDeblock", "PR2Flush", "SMXFlush", "End"] #List of non-pairedend steps
    solNames={'1':'Incorporation Mix', '2':'diH20', '3':'Scan Mix', '4':'High Salt Buffer', '5':'Incorporation Buffer',
              '6':'Cleavage Mix', '7':'Cleavage Buffer', '10':'Resynthesis Mix', '11':'Linerization Mix', '12':'Block Mix',
              '13':'Amplification Mix', '14':'Preamplification Mix', '15':'Formamide', '16':'Read 2 primer mix',
              '19':'Dil NaOH', '21':'NaOH', '21':'Wash buffer', '22':'Index primer mix'}
    
    for z in w:
        deadVolume[z]=55
    def __init__(self, xml, lanes, addpulse, reagEst, extraLane, output=None):
        self.xml=xml
        self.recipe=ET.parse(xml, parser=CommentedTreeBuilder())
        self.lanes=lanes
        self.addpulse=addpulse
        self.extraLane=extraLane
        self.output=output
        self.reagEst=reagEst

    ##Finds new solution volumes
    def newVolumes(self, IV, x):
        return int(IV)+int(round(self.deadVolume[x]*((8.0/len(self.lanes))-1)))
    ##Finds new solution volume for 1 lane scenario
    def specVolumes(self, IV, x):
        return int(IV)+self.deadVolume[x]*3
    ##Undoes calculation to find actual volume used
    ##Converts from 10 microliters to 1 mL
    def reverseVolumes(self, CV):
        return int(CV)*(len(self.lanes)/800.0)

    ##Adjusts solution volumes, checks for a read 2 portion and adds pulse if needed
    def newRecipe(self):
        chem=self.recipe.findall("ChemistryDefinitions/*")
        y=0
        pEnd=False            
        for s in chem:
            x=None
            if s.tag==ET.Comment:
                if 'RESYNTHESIS' in s.text:
                    y=len(self.lanes)
                    pEnd=True
                else:
                    y=0
            elif s.get('Name')!='Resynthesis':
                z=0 #Index of steps in each Chemistry
                steps=s.getiterator()
                for action in steps:
                    if action.tag=="PumpToFlowcell":
                        IV=action.get("Volume")
                        if x!=action.get("Solution"):
                            x=action.get("Solution")
                            if y==1 and self.extraLane:
                                sv=self.specVolumes(IV,x)
                            else:
                                sv=self.newVolumes(IV, x)
                            if self.addpulse:
                                if s.get('Name') in self.chemNames:
                                    mix=ET.Element("PumpToFlowcell", attrib=action.attrib)
                                    mix.set("Volume", "20")
                                    sv+=40
                                    s.insert(z,ET.Element("Wait", attrib={"Duration":"5000"}))
                                    z+=1
                                    s.insert(z,mix)
                                    z+=1
                            action.set("Volume", str(sv))
                    z+=1
            elif len(self.lanes)!=8:
                ptfc=s.getiterator("PumpToFlowcell")
                for sol in ptfc:
                    if sol.get("Solution")=="13":
                        sol.set("Volume", "54")
                    else:
                        sol.set("Volume", "42")
        return pEnd
    
    ##Finds time needed to run
    def timeEstimater(self, l):
        chemtimes={'Ruu':0}
        chemsteps=self.recipe.findall("ChemistryDefinitions/Chemistry")
        for step in chemsteps:
            time=0
            for s in step.findall("*"):
                if s.tag=="PumpToFlowcell":
                    time+=float(s.get("Volume"))/float(s.get("AspirationRate"))
                else:
                    time+=float(s.get("Duration",default=0))/60000
            if 'Prime' in step.get('Name'):
                chemtimes[step.get('Name')]=1.5
            else:
                chemtimes[step.get('Name')]=time
        totalTime=0
        for p in self.recipe.findall("Protocol/ChemistryRef"):
            totalTime+=chemtimes[p.get("Name",default='Ruu')]
        for p in self.recipe.findall("Protocol/Incorporation"):
            totalTime+=chemtimes[p.get("ChemistryName",default='Ruu')]
            totalTime+=5*len(self.lanes)
        #print chemtimes
        tEst = 'Flowcell will take approximately '+ str(int(round(totalTime/60)))+ ' hours to run. '
        ##Only necessary if there is a turnaround
        if l:
            turnaround=0
            cRef=self.recipe.findall('Protocol/ChemistryRef')
            d=cRef.index(self.recipe.find("Protocol/ChemistryRef/[@Name='Denaturation']"))
            r=cRef.index(self.recipe.find("Protocol/ChemistryRef/[@Name='Rehybridisation']"))+1
            for tas in cRef[d:r]:
                turnaround+=chemtimes[tas.get('Name', default='Ruu')]
            turnaround+=15
            tEst=tEst+'The turnaround will take about '\
                  +str(int(turnaround/60)) +' hours and '+str(int(round(turnaround%60,-1)))+' minutes. '
        return tEst

    ##Finds reagent use
    def reagentEstimater(self, elements, basevol):
        totalvol={'www':0}
        for p in elements:
            dicOfsol={}
            if p.tag=="ChemistryRef":
                dicOfsol=basevol[p.get('Name',default='Bre')]
            elif p.tag=="Incorporation":
                dicOfsol=basevol[p.get('ChemistryName',default='Bre')]
            for solution, volume in dicOfsol.items():
                totalvol[solution]=totalvol.setdefault(solution,0)+volume               
        del totalvol['www']
        return totalvol


    ##adds rinse with water if necessary
    ##actually is always necessary regardless of pulse
    def turnaroundPulse(self, r2):
        if r2:
            ChemistryDef=self.recipe.find("ChemistryDefinitions")
            RinseStep=ET.Element("Chemistry", attrib={"Name":"WaterRinse"})
            ET.SubElement(RinseStep, "PumpToFlowcell", attrib={"Solution":"2", "AspirationRate":"250", \
                                                                 "DispenseRate":"2500", "Volume":"200"})
            ChemistryDef.append(RinseStep)
            protocol=self.recipe.find("Protocol")
            rehybridise=protocol.find("ChemistryRef/[@Name='Rehybridisation']")#finding rehybridization step
            ri=protocol.getchildren().index(rehybridise)
            reconnect=ET.Element("UserWait", attrib={"Message":"Please reconnect all syringes for the water rinse."})
            rinse=ET.Element("ChemistryRef", attrib={"Name":"WaterRinse"})
            disconnect=ET.Element("UserWait", attrib={"Message":"Please disconnect syringes not attached to active lanes."})
            protocol.insert(ri+1,reconnect)
            protocol.insert(ri+2,rinse)
            protocol.insert(ri+3,disconnect)

    def newXML(self):
        r2=self.newRecipe()
        #self.turnaroundPulse(r2)
        if self.output!=None:
            root=self.output
        else:
            root=os.path.dirname(self.xml)
        tiles=self.recipe.find("TileSelection").find("Incorporation")
        LI=tiles.getiterator("Lane")
        for qb in LI:
            if qb.get("Index") not in self.lanes[:]:
                tiles.remove(qb)
        x=self.lanes[0]
        for l in range(1,len(self.lanes)):
            x=x+', '+self.lanes[l]
        prefix="Partial_"+str(len(self.lanes))+"lanes_"+x+"_"

        ChemistryDef=self.recipe.find("ChemistryDefinitions")
        RinseStep=ET.Element("Chemistry", attrib={"Name":"WaterRinse"})
        ET.SubElement(RinseStep, "PumpToFlowcell", attrib={"Solution":"2", "AspirationRate":"250", \
                                                             "DispenseRate":"2500", "Volume":"200"})
        ChemistryDef.append(RinseStep)
        protocol=self.recipe.find("Protocol")
        reads=self.readFinder(r2)

        ##Things that need to happen in read1
        r1messages=[m for m in reads['read1'] if m.tag == 'UserWait']
        message=r1messages[0]
        message1_index = protocol.getchildren().index(message)
        
        reconnect=ET.Element("UserWait", attrib={"Message":"Please make sure all syringes are connected. A water rinse will now be performed"})
        rinse=ET.Element("ChemistryRef", attrib={"Name":"WaterRinse"})
        disconnect=ET.Element("UserWait", attrib={"Message":"Please disconnect all syringes except for those attached to lanes " + x})
        protocol.insert(message1_index+1, reconnect)
        protocol.insert(message1_index+2, rinse)
        protocol.insert(message1_index+3, disconnect)

        discwarn=" The run is a " + str(len(self.lanes)) + "-lane partial flowcell run on lanes " + x + ". "
        (a,b,c)=message.get('Message').partition('.')

        if r2:
        ##Things that need to happen in the turnaround
            turnmessages = [m for m in reads['turnaround'] if m.tag == 'UserWait']
            
        ##Things that need to happen in read2
            ChemistryDef=self.recipe.find("ChemistryDefinitions")
            RinseStep=ET.Element("Chemistry", attrib={"Name":"WaterRinse"})
            ET.SubElement(RinseStep, "PumpToFlowcell", attrib={"Solution":"2", "AspirationRate":"250", \
                                                                 "DispenseRate":"2500", "Volume":"200"})
            ChemistryDef.append(RinseStep)
            rehybridise=protocol.find("ChemistryRef/[@Name='Rehybridisation']")#finding rehybridization step
            ri=protocol.getchildren().index(rehybridise)
            reconnect=ET.Element("UserWait", attrib={"Message":"Please reconnect all syringes for the water rinse."})
            protocol.insert(ri+1,reconnect)
            protocol.insert(ri+2,rinse)
            protocol.insert(ri+3,disconnect)
            
            r2messages = [m for m in reads['read2'] if m.tag=='UserWait']
            if self.addpulse:
                prefix=prefix+"WithPulse_"
                discwarn = discwarn + "An extended pulse has been added to increase quality for long runs. "
            if self.extraLane and len(self.lanes)==1:
                (first, period, last)=r1messages[len(r1messages)-1].get('Message').partition('.')
                r1messages[len(r1messages)-1].set("Message",first+'. Please attach extra pump for the turnaround.'+last)
                (first, period, last)=turnmessages[0].get('Message').partition('.')
                turnmessages[0].set("Message",first+'. Make sure the extra turnaround pump is disconnected.'+last)
                discwarn = discwarn + "During the turnaround, an extra lane will be used to reduce run time to reasonable levels. "\
                           + "This will increase reagent use and essentially burn a lane. If the extra lane is not wanted, "\
                           + "please generate a new recipe file with the --noextrapump flag turned on. "
        r1messages[0].set("Message",a+b+discwarn+c)
       
        if self.reagEst:
            reagentext=open(os.path.join(root, prefix + "Estimate.txt"), 'w')
            reagentext.write(self.timeEstimater(r2)+'\n\n')
            basevol={'Bre':{}}
            chemsteps=self.recipe.findall("ChemistryDefinitions/Chemistry")
            for step in chemsteps:
                chemname=step.get('Name')
                basevol[chemname]={}
                solutionlist=basevol[chemname]
                for s in step.findall("*"):
                    solutionlist[s.get('Solution',default='www')]=solutionlist.setdefault(s.get('Solution',default='www'),0)\
                                                                      +self.reverseVolumes(float(s.get('Volume',default=0)))
            read1vol=self.reagentEstimater(reads['read1'], basevol)
            reagentext.write('READ 1 VOLUMES/INDEX (IF APPLICABLE)'.center(40))
            reagentext.write('\n')
            reagentext.write('{0:23}{1}\n'.format('Reagent', 'Volume (mL)'))
            for solution, volume in read1vol.items():
                reagentext.write('{0:23}{1}\n'.format(self.solNames[solution], round(volume, 1)))
            if r2:
                turnavol=self.reagentEstimater(reads['turnaround'], basevol)
                reagentext.write('\n')
                reagentext.write('TURNAROUND VOLUMES'.center(40))
                reagentext.write('\n')
                reagentext.write('{0:23}{1}\n'.format('Reagent', 'Volume (mL)'))
                for solution, volume in turnavol.items():
                    if solution=='21':
                        volume+=1
                    else:
                        volume+=.65
                    reagentext.write('{0:23}{1}\n'.format(self.solNames[solution], round(volume, 1)))
                read2vol=self.reagentEstimater(reads['read2'], basevol)
                reagentext.write('\n')
                reagentext.write('READ 2 VOLUMES'.center(40))
                reagentext.write('\n')
                reagentext.write('{0:23}{1}\n'.format('Reagent', 'Volume (mL)'))
                for solution, volume in read2vol.items():
                    reagentext.write('{0:23}{1}\n'.format(self.solNames[solution], round(volume, 1)))
        self.recipe.write(file_prefixer(self.xml,root,prefix))

    def readFinder(self, r2):
        ##finds first read location
        proslice={}
        protocol=self.recipe.findall("Protocol/*")
        for x in protocol:
            turnaroundindex = len(proslice)-1
            if x.tag==ET.Comment and "READ 1 SBS" in x.text:
                read1index = protocol.index(x)
            else:
                read1index = 0
        if r2:
            for q in protocol:
                if q.tag==ET.Comment and "PREPARATION / RESYNTHESIS" in q.text:
                    turnaroundindex=protocol.index(q)
                elif q.tag==ET.Comment and "READ 2 SBS" in q.text:
                    read2index=protocol.index(q)
            proslice['read1']=protocol[read1index:turnaroundindex]
            proslice['turnaround']=protocol[turnaroundindex:read2index]
            proslice['read2']=protocol[read2index:]
        else:
            proslice['read1']=protocol[read1index:]
        return proslice
            
            
            

parser=argparse.ArgumentParser(description='Adjusts the recipe to account for number of lanes used')
parser.add_argument('xml', help='the recipe file')
parser.add_argument('--pulseon', action='store_true', help = 'adds extended extended pulse for long runs [DEFAULT:FALSE]')
parser.add_argument('--output', help = 'changes the output directory [DEFAULT:INPUT DIRECTORY]')
parser.add_argument('--estimate', action='store_true', help = 'outputs estimated volume usage in Reagent_estimate.txt [DEFAULT:FALSE]')
parser.add_argument('--noextrapump', action='store_false', help = 'removes additional lane added for single lane flowcell turnaround [DEFAULT:FALSE]')
parser.add_argument('lanes', nargs='+', type=str, choices=['1','2','3','4','5','6','7','8'], help='all lanes being used')
args=parser.parse_args()
recipeMaker(args.xml,args.lanes,args.pulseon, args.estimate, args.noextrapump, args.output).newXML()
        

